/*
 * Copyright (C) 2005 by Robert James Kaes <rjk@wormbytes.ca>
 *   All rights reserved.
 *
 * This code is released into the public domain.  Enjoy.
 *
 * PURPOSE:
 * Generates all possible combinations for a mechanical push button
 * lock like those found in offices and labs in the 1990's.  Locks
 * can have up to eight buttons.  In addition, multi-button pairs
 * are allowed.  For a complete description of what the program is
 * required to accomplish see:
 *   
 * http://www.sentex.net/~mwandel/careers/sample_problem.html
 * 
 */

#include <stdio.h>
#include <stdlib.h>


/* Max number of keys allowed in the combination lock */
#define MAX_KEYS 8

/*
 * Store the pre-allocated button presses.  Since the problem domain restricts
 * the number of buttons to 8, there can only be a max of 36 potential presses.
 */
struct buttons_s {
        int size;
        unsigned char buttons[MAX_KEYS + ((MAX_KEYS * (MAX_KEYS - 1)) / 2)];
} g_buttons;


/* Store the points to the various offsets into the button_s.button structure */
static int g_offsets[MAX_KEYS];


/*
 * Generate a list of all the possible single and pair button combinations.
 */
static void generate_button_presses(int limit)
{
        int i, j;
        unsigned char *p = g_buttons.buttons;

        /* add the total single buttons to the total pair buttons */
        g_buttons.size = limit + ((limit * (limit - 1)) / 2);

        /* add the single buttons first */
        for (i = 1; i <= limit; ++i)
                *p++ = i;

        /* add the buttons pairs */
        for (i = 1; i <= limit - 1; ++i)
                for (j = i + 1; j <= limit; ++j)
                        *p++ = ((i << 4) | j);
}


/*
 * Initialize the combination array for the combinatoric k-of-n.  See the
 * description for next_choice() below for an explanation of where
 * the function is derived from.
 *
 * Parameters:
 *  v - array containing the combinations generated
 *  n - value of maximum number in set
 *  k - number of values to pull from the set
 */
static void init_choice_sequence(int *v, int n, int k)
{
        int i;
        for (i = 0; i != k; ++i)
                v[i] = i;
        v[k] = n + 1;
}


/*
 * Choices the next combination of k from the set n.  The algorithm
 * is from "Constructive Combinatorics," Stanton and White.  However,
 * I do not have the book where this algorithm is described.  I found
 * a description of the algorithm online.
 *
 * Parameters:
 *  v - array containing the current last combination
 *  n - value of the max number in the set
 *  k - how many values to pull from the set
 *
 * Returns:
 *  0 - no more choices available
 *  1 - another choice _is_ available
 */
static int next_choice(int *v, int n, int k)
{
        int i, j;

        if (v[0] >= n - k + 1) return 0;

        j = 0;
        while (v[j + 1] <= v[j] + 1) { ++j; }

        v[j]++;

        for (i = 0; i != j; ++i) v[i] = i;

        return 1;
}


/*
 * Swap the contents of two memory locations.  Uses the XOR swap
 * technique to avoid using another RAM address.
 *
 * NOTE: Using XOR is possible in this application because the
 * values of the keys is limited to integers between 1 and 8.  In
 * general, however, it is better to use a temporary variable during
 * the swap.  For RAM constrained platforms, the XOR solution saves
 * a byte.
 *
 * Parameters:
 *  i - pointer to first value to swap
 *  j - pointer to second value to swap
 */
static inline void swap_elements(int *i, int *j)
{
        *j ^= *i;
        *i ^= *j;
        *j ^= *i;
}


/*
 * Reverse all elements between first and last; where last points
 * to one past the end of the array (like in the STL.)
 *
 * Parameters:
 *  first - pointer to beginning of sequence
 *  last - one _past_ the end of the sequence
 */
static void reverse_elements(int *first, int *last)
{
        if (first + 1 == last) return;

        --last;
        while (first < last)
                swap_elements(first++, last--);
}


/*
 * Generates the next permutation of the sequence defined by first..last.
 * Works identically to the next_permutation() function in the C++ STL.
 *
 * Prerequisites:
 *  The sequence first..last must be lexically sorted.
 *
 * Parameters:
 *  first - pointer to beginning of sequence of bytes
 *  last - one _past_ the end of the sequence.
 *
 * Returns:
 *  0 - no more permutations possible
 *  1 - another permutation is possible.
 */
static int next_permutation(int *first, int *last)
{
        int *i, *j, *k;

        /* Nothing to do for a one-element list */
        if (first + 1 == last) return 0;

        /*
         * i and j are adjacent pointers (with i preceding j.)  The
         * pointers begin at the end of the range and work toward the

         * beginning.
         */
        i = last - 1;
        for (;;) {
                j = i--;
                
                if (*i < *j) {
                        k = last;
                        while (*i > *--k) {}
                        swap_elements(i, k);
                        reverse_elements(j, last);
                        return 1;
                }
                
                if (i == first) {
                        reverse_elements(first, last);
                        return 0;
                }
        }
}


/*
 * Displays all the permutations of an array of size bytes.
 *
 * Parameters:
 *  array - pointer to beginning of the array of bytes
 *  size - size of the array.
 */
static void print_permutations(int *array, int size)
{
        int i, v;
        
        do {
                for (i = 0; i != size; ++i) {
                        v = g_buttons.buttons[array[i]];
                        if (v & 0xf0) printf("%d-", (v >> 4) & 0xf);
                        printf("%d ", v & 0xf);
                }
                putchar('\n');
        } while (next_permutation(array, array + size));
}


/*
 * Ensure that the combination is valid (no duplicated keys.)  Returns a
 * boolean.
 */
static int valid_combination(int *array, int size)
{
        unsigned char total_mask = 0;
        unsigned char byte, mask;
        int i;

        for (i = 0; i != size; ++i) {
                byte = g_buttons.buttons[g_offsets[i]];
                mask = 1 << ((byte & 0x0f) - 1);
                if (byte & 0xf0)
                        mask |= 1 << (((byte & 0xf0) >> 4) - 1);

                if (total_mask & mask)
                        return 0;
                total_mask |= mask;
        }

        return 1;
}


/*
 * Go through all the combination sequences and then print all the
 * permutations of that combination.
 */
static void print_sequences(int keys)
{
        int i;

        for (i = 1; i <= keys; ++i) {
                init_choice_sequence(g_offsets, g_buttons.size - 1, i);
                do {
                        if (valid_combination(g_offsets, i))
                                print_permutations(g_offsets, i);
                } while (next_choice(g_offsets, g_buttons.size - 1, i));
        }
}


/*
 * Driver for the rest of the program.  Accepts the number of keys
 * from the command line and validates the value.
 *
 * Return:
 *  0 - success
 *  1 - usage errors
 */
int main(int argc, char **argv)
{
        int keys;
        
        if (--argc < 1) {
                fprintf(stderr, "Usage: lock keys\n");
                return 1;
        }

        keys = atoi(argv[1]);
        if (keys < 1 || keys > MAX_KEYS) {
                fprintf(stderr, "Keys outside of range (1..%d)\n", MAX_KEYS);
                return 1;
        }

        /*
         * Generate the initial pool of possible entries.  These include
         * both the single button and pairs.
         */
        generate_button_presses(keys);

        /* Display the various combinations. */
        print_sequences(keys);

        return 0;
}
